'' MPE_SPI_GOLEM_002
'' A GameOfLife for the DOGM 4-PIN 2008-01 by deSilva

'  A most simple implementation, the status is kept in bytes
'  in a "lines" x 128 array ("lines" are 32 or 64 resp.)
'  It can be experimented with "divide and conquer"; computing
'  the next genereation is split to another, or 2, or even 4 COGs

'  It shows that the SPIN implementation of SPI is NOT a bottleneck
'  for this kind of algorithms 

  _clkmode        = xtal1 + pll8x     ' change!
  _xinfreq        = 10_000_000        ' change!

  LINES = 32     ' change to 64 for full display
  PARTS =  4     ' number of additional COGs working in parallel (1, 2 or 4)
                 ' -1 (rather than expected "0") indicates NO parallelism
  ledPin = 6     ' idle indicator

OBJ
  gdogm : "MPE_SPI_DOGM_005"

VAR
  BYTE complete[||PARTS]       ' feed back of COGs when completed
  BYTE GOL[128*lines]          ' new generation is assembled here
  BYTE upperBorder[128]        ' border, just zeros
  BYTE GOLB[128*lines]         ' old generation to read and to display from
  BYTE lowerBorder[128]

PUB golem | x, i, completed, part, aStack[20*||PARTS]

  DIRA[ledPin]~~                       ' will indicate main COG idle

  gdogm.init(0)

  ' slow wipeout
  REPEAT i FROM 0 TO constant(LINES/8-1) 
     gdogm.clearPage(i)

  X:=CNT
  REPEAT i FROM 0 to constant(128*LINES-1)
     GOL[i] := (?X)&1

    
  REPEAT
      
      LONGMOVE(@GOLB, @GOL, constant(128/4*LINES))  

      IF PARTS<1                       ' no parallel COGS!
         nextGeneration(0, @complete)
      ELSE
         REPEAT part from 0 to PARTS-1 ' recruit some COGs for the dirty work
           cognew(                                                         {
}              nextGeneration(part*constant(128/||PARTS), @complete+part), {
}              @aStack+20*4*part)

      displayGOL                       ' displays old GOLB!
      
      outa[ledPin]~~

      REPEAT                            ' wait for all COGs ready
          completed~   
          REPEAT part FROM 0 to ||PARTS-1
             completed += complete[part]
      UNTIL completed =>||PARTS     
      outa[ledPin]~

PRI nextGeneration(offset, completePointer) |i,c
      BYTE[completePointer]~
       
      longfill(@GOL+offset*lines,0, constant(128/4/||PARTS)*lines) 
      REPEAT i FROM offset*lines to (offset+constant(128/||PARTS))*lines-1
        c := GOLB[i-1]+GOLB[i+1]+GOLB[i-lines]+GOLB[i+lines]+    {
}            GOLB[i+1+lines]+GOLB[i-1+lines]+GOLB[i+1-lines]+GOLB[i-1-lines]
        IF c == 3 or (C==2 and GOLB[i])
           GOL[i] := 1
           
      BYTE[completePointer]~~     


PRI displayGOL |page, posString,col, c, cc[128/4],refAddr

  posString.byte[2] := $00
  posString.byte[1] := $10
  REPEAT page FROM 0 TO constant(lines/8)-1

     posString.byte[0] := $B0 + page
     gdogm.write(@posString,3,0)
     REPEAT col FROM 0 to 127
        c :=0
        refAddr := col*lines+page*8
        REPEAT 8          
            c := (c<<1) + GOLB[refAddr++]
        BYTE[@cc+col] := c><8
     gdogm.write(@cc, 128, 1)
     
'=== The End